"""String Validation Methods.
"""


def validate_name(argument: str) -> bool:
    """ Determine whether an argument is a non-empty string.
- Does not count whitespace.
- Uses the strip method to remove empty space.

**Parameters:**
 - argument (str) : The given argument.

**Returns:**
 bool - True if the argument qualifies as valid.
    """
    if argument is None or not isinstance(argument, str):
        return False
    elif len(argument.strip()) < 1:
        return False
    return True


""" Changelist Init Package.
"""
from typing import Iterable

from changelist_data import ChangelistDataStorage
from changelist_data.changelist import Changelist, get_default_cl
from changelist_data.file_change import FileChange

from changelist_init import git, fc_to_cl_map
from changelist_init.input.input_data import InputData


_DEFAULT_CHANGELIST_ID = '12345678'
_DEFAULT_CHANGELIST_NAME = "Initial Changelist"


def initialize_file_changes(
    input_data: InputData,
) -> list[FileChange]:
    """ Get up-to-date File Change information in a list.
 - Initializing Changelists begins by requesting File Information from Git.

**Parameters:**
 - input_data (InputData): ChangelistInit Input Package Data object, containing program input.

**Returns:**
 list[FileChange] - The List of File Changes, freshly squeezed from git.
    """
    return list(git.generate_file_changes(input_data.include_untracked))


def merge_file_changes(
    existing_lists: list[Changelist],
    files: Iterable[FileChange],
) -> bool:
    """ Merge FileChange into Changelists.
 - Leaves existing files in their Changelists.
 - Inserts all new files into the default Changelist.
 - Creates InitialChangelist with id:12345678 containing all files, if existing lists empty.

**Parameters:**
 - existing_lists (list[Changelist]): The list of Changelists that the InputData is starting with.
 - files (list[FileChange]): The new FileChanges that have been obtained from Git.

**Returns:**
 bool - True after the operation has finished.
    """
    if (default_cl := get_default_cl(existing_lists)) is not None:
        default_cl.changes.extend(
            fc_to_cl_map.merge_fc_generator(
                changelists=existing_lists,
                file_changes=files,
            )
        )
    else:
        existing_lists.append(Changelist(
            id=_DEFAULT_CHANGELIST_ID,
            name=_DEFAULT_CHANGELIST_NAME,
            changes=files if isinstance(files, list) else list(files),
            is_default=True,
        ))
    return True


def init_storage(
    storage: ChangelistDataStorage,
    include_untracked: bool,
) -> bool:
    """ Get New FileChange Information, Merge into Changelists Data.

**Parameters:*
 - storage (ChangelistDataStorage): The Storage object to obtain existing CL from, and send updates to.
 - include_untracked (bool): Whether to tell git to include untracked files.

**Returns:**
 bool - True if the initialized data merged into Changelists Storage object successfully.
    """
    if not merge_file_changes(
        cl := storage.get_changelists(),
        git.generate_file_changes(include_untracked)
    ):
        return False
    storage.update_changelists(cl)
    return True # Successful Merge


""" Maps Git Status data into FileChange data.
"""
from itertools import groupby
from typing import Callable, Iterable, Generator

from changelist_data import file_change
from changelist_data.file_change import FileChange

from changelist_init.git.git_file_status import GitFileStatus


def map_file_status_to_changes(
    git_files: Iterable[GitFileStatus],
) -> Generator[FileChange, None, None]:
    """ Categorize by Status Code, and Map to FileChange data objects.

    Parameters:
    - git_files (Iterable[GitFileStatus]): An iterable or Generator providing GitFileStatus objects.

    Returns:
    FileChange - Yield by Generator.
    """
    for code, group in groupby(git_files, lambda w: w.code):
        mapping_function = get_status_code_change_map(code)
        for file_status in group:
            yield mapping_function(
                map_status_path_to_change(file_status.file_path)
            )


def get_status_code_change_map(
    code: str,
) -> Callable[[str, ], FileChange]:
    """ Get a FileChange mapping callable for a specific code.

    Parameters:
    - code (str): The status code, determining what kind of FileChange (create, modify, delete)

    Returns:
    Callable[str, FileChange] - A function that maps a FileChange path into the FileChange object.
    """
    if code in ('M ', ' M', 'MM'):
        return file_change.update_fc
    if code in ('A ', ' A', 'AM', 'MA'):
        return file_change.create_fc
    if code in ('D ', ' D', 'MD', 'DM'):
        return file_change.delete_fc
    if '?' in code or '!' in code:
        return file_change.create_fc
    exit(f"Unknown Code: {code}")


def map_status_path_to_change(
    status_path: str,
) -> str:
    """ Convert Status File path to FileChange path.
        Adds a leading slash character.
    """
    return '/' + status_path if not status_path.startswith('/') else status_path


""" Methods for Mapping FileChanges into Changelist Data structures.
"""
from typing import Iterable, Generator

from changelist_data import Changelist
from changelist_data.file_change import FileChange


def create_fc_to_cl_dict(
    changelists: list[Changelist],
) -> dict[str, Changelist]:
    """ Initialize the Map of Existing FileChanges.

**Parameters:
 - changelists (list[Changelist]): The changelists to be inserted into the map.

**Returns:**
 dict[str, Changelist] - A map from file path to the Changelist object that contains it.
    """
    cl_map: dict[str, Changelist] = {}
    for cl in changelists:
        for fc in cl.changes:
            if (before := fc.before_path) is not None:
                cl_map[before] = cl
            if (after := fc.after_path) is not None:
                cl_map[after] = cl
    return cl_map


def offer_fc_to_cl_dict(
    fc_map: dict[str, Changelist],
    file: FileChange
) -> bool:
    """ Map a FileChange into an existing Changelist.
 - Searches the Dict for both FileChange paths, starting with the before_path.
 - Appends FC to CL changes and returns true if found.

**Parameters:**
 - fc_map (dict[str, Changelist]): The FileChange-to-Changelist Map.
 - file (FileChange): The FC to search the dict for, and add to the Changelist if found.

**Returns:**
 bool - True if the FC was added to a CL. False if there was no match with the dict.
    """
    if (before := file.before_path) is not None:
        if (cl := fc_map.get(before)) is not None:  # Match!
            cl.changes.append(file)
            return True
    elif (after := file.after_path) is not None:
        if (cl := fc_map.get(after)) is not None:  # Match!
            cl.changes.append(file)
            return True
    else:
        exit("This FC has no file path property.")
    return False  # FC not in map


def merge_fc_generator(
    changelists: list[Changelist],
    file_changes: Iterable[FileChange],
) -> Generator[FileChange, None, None]:
    """ Merge FC into existing CL, applying the fc_to_cl_map module.

**Parameters:**
 - existing_lists (list[Changelist]): The list of existing Changelists.
 - files (list[FileChange]): The list of FileChange objects produced during initialization.

**Yields:**
 Generator[FileChange] - The FileChange objects that are new, not present in existing Changelists.
    """
    cl_map = create_fc_to_cl_dict(changelists)
    # Clear Existing Lists before merging new Files
    for cl in changelists:
        cl.changes.clear()
    # Search for matches using map
    for fc in file_changes:
        if not offer_fc_to_cl_dict(cl_map, fc):
            yield fc


""" Runner for Git Status Operation
"""
import subprocess


def run_git_status() -> str:
    """ Run a Git Status Process and Return the Output.

    Returns:
    str - The output of the Git Status Operation.
    """
    result = subprocess.run(
        args=['git', '--no-optional-locks', 'status', '--porcelain', '--no-renames'],
        capture_output=True,
        text=True,
        universal_newlines=True,
        shell=False,
        timeout=5,
    )
    if (error := result.stderr) is not None and not len(error) < 1:
        exit(f"Git Status Runner Error: {error}")
    return result.stdout


def run_untracked_status() -> str:
    """ Configure git for obtaining untracked files, then reset.
        - Uses multiple git commands to achieve the desired result.
        - Runs a Git Add command, then a soft reset.
    """
    # Add All Untracked Paths without staging
    git_add_output = subprocess.run(
        args=['git', 'add', '-N', '.'],
        capture_output=True,
        text=True,
        universal_newlines=True,
        shell=False,
        timeout=3,
    )
    if (error := git_add_output.stderr) is not None and not len(error) < 1:
        exit(f"Git Add Error: {error}")
    #
    result = run_git_status()
    # Before returning result, reset git to keep staging area clean
    # Use a soft reset, instead of default (mixed)
    git_reset_output = subprocess.run(
        args=['git', 'reset', '-q', '--soft'],
        capture_output=True,
        text=True,
        universal_newlines=True,
        shell=False,
        timeout=3,
    )
    if (error := git_reset_output.stderr) is not None and not len(error) < 1:
        exit(f"Git Reset Error: {error}")
    return result



""" Input Sub-Package Methods.
"""
from changelist_data import load_storage_from_file_arguments

from changelist_init.input.argument_parser import parse_arguments
from changelist_init.input.input_data import InputData


def validate_input(
    arguments: list[str],
) -> InputData:
    """ Parse and Validate the Arguments, and return Input Data.

**Parameters:**
 - arguments (list[str]): The arguments received by the program.

**Returns:**
 InputData - The InputData containing the program inputs. The other packages will process the data from here.
    """
    arg_data = parse_arguments(arguments)
    return InputData(
        storage=load_storage_from_file_arguments(arg_data.changelists_file, arg_data.workspace_file),
        include_untracked=arg_data.include_untracked,
    )



""" The Arguments Received from the Command Line Input.
- This DataClass is created after the argument syntax is validated.
"""
from dataclasses import dataclass


@dataclass(frozen=True)
class ArgumentData:
    """ The syntactically valid arguments received by the Program.

**Fields:**
 - changelists_file (str?): The string path to the Changelists Data File.
 - workspace_file (str?): The string path to the Workspace File.
 - include_untracked (bool): Whether to include untracked files.
 - generate_sort_xml (bool): Generate the config.xml file for the project..
    """
    changelists_file: str | None = None
    workspace_file: str | None = None
    include_untracked: bool = False
    generate_sort_xml: bool = False


""" Valid Input Data Class.
"""
from dataclasses import dataclass

from changelist_data.storage import ChangelistDataStorage


@dataclass(frozen=True)
class InputData:
    """ A Data Class Containing Program Input.

**Fields:**
 - storage (ChangelistDataStorage): The Storage object used for Data IO.
 - include_untracked (bool): Whether to include untracked files.
    """
    storage: ChangelistDataStorage
    include_untracked: bool = False



""" Container for the Git Status Information.
    Files in Git Status are categorized into 3 groups.
    1. Staged
    2. Unstaged
    3. Untracked
    These three groups are stored in three lists in this data class.
"""
from dataclasses import dataclass, field
from typing import Generator

from changelist_init.git.git_file_status import GitFileStatus
from changelist_init.git.git_tracking_status import GitTrackingStatus


@dataclass(frozen=True)
class GitStatusLists:
    """ The Lists summarizing the current Git Status.
    """
    _lists: dict[GitTrackingStatus, list[GitFileStatus]] = field(default_factory=lambda: {})

    def get_list(
        self, tracking_status: GitTrackingStatus
    ) -> list[GitFileStatus] | None:
        """ Obtain a List of File Status objects with the given Tracking Status.
        """
        return self._lists.get(tracking_status)

    def add_file_status(
        self, file_status: GitFileStatus | None,
    ) -> bool:
        """ Add A File Status to the Lists.
        """
        status = file_status.get_tracking_status()
        if (initial_list := self.get_list(status)) is not None:
            initial_list.append(file_status)
            return True
        self._lists[status] = [file_status]
        return True

    def merge_tracked(self) -> Generator[GitFileStatus, None, None]:
        """ Combine the Tracked Lists.
        These are currently divided into 3 categories: Staged, Unstaged, and PartialStage
        """
        tracked_status = (GitTrackingStatus.STAGED, GitTrackingStatus.UNSTAGED, GitTrackingStatus.PARTIAL_STAGE)
        for e in tracked_status:
            if (tracked_list := self.get_list(e)) is not None:
                yield from tracked_list

    def merge_all(self) -> Generator[GitFileStatus, None, None]:
        """ Combine the Tracked Lists.
        These are currently divided into 3 categories: Staged, Unstaged, and PartialStage
        """
        for e in GitTrackingStatus:
            if (file_list := self.get_list(e)) is not None:
                for f in file_list:
                    yield f



""" Defines and Validates Argument Syntax.
 - Encapsulates Argument Parser.
 - Returns Argument Data, the args provided by the User.
"""
from argparse import ArgumentParser
from sys import exit
from typing import Optional

from changelist_init.input.argument_data import ArgumentData
from changelist_init.input.string_validation import validate_name


def parse_arguments(
    args: Optional[list[str]] = None
) -> ArgumentData:
    """ Parse command line arguments.

**Parameters:**
 - args: A list of argument strings.

**Returns:**
 ArgumentData : Container for Valid Argument Data.
    """
    if args is None or len(args) == 0:
        return ArgumentData()
    # Initialize the Parser and Parse Immediately
    try:
        parsed_args = _define_arguments().parse_args(args)
    except SystemExit:
        exit("Unable to Parse Arguments.")
    return _validate_arguments(parsed_args)


def _validate_arguments(
    parsed_args,
) -> ArgumentData:
    """ Checks the values received from the ArgParser.
- Uses Validate Name method from StringValidation.

**Parameters:**
 - parsed_args : The object returned by ArgumentParser.

**Returns:**
 ArgumentData - A DataClass of syntactically correct arguments.
    """
    if (changelists_file := parsed_args.changelists_file) is not None:
        if not validate_name(changelists_file):
            exit("The Changelists File name was invalid.")
    if (workspace_file := parsed_args.workspace_file) is not None:
        if not validate_name(workspace_file):
            exit("The Workspace File name was invalid.")
    return ArgumentData(
        changelists_file=parsed_args.changelists_file,
        workspace_file=parsed_args.workspace_file,
        include_untracked=parsed_args.include_untracked,
    )


def _define_arguments() -> ArgumentParser:
    """ Initializes and Defines Argument Parser.
 - Sets Required/Optional Arguments and Flags.

**Returns:**
 argparse.ArgumentParser - An instance with all supported Arguments.
    """
    parser = ArgumentParser(
        description="Initializes and updates the Changelist data storage file with git status information.",
    )
    parser.color = True
    # Optional Arguments
    parser.add_argument(
        '--changelists_file',
        type=str,
        default=None,
        help='The Path to the Changelists Data File. Searches default paths if none.',
    )
    parser.add_argument(
        '--workspace_file',
        type=str,
        default=None,
        help='The Path to the Workspace Data File. Searches default paths if none.',
    )
    parser.add_argument(
        "--include_untracked", "-u",
        action='store_true',
        default=False,
        help='The option to include untracked files in changelists.',
    )
    return parser



""" Git Management Package.
"""
from typing import Generator

from changelist_data.file_change import FileChange

from changelist_init.git import status_runner, status_reader, status_change_mapping
from changelist_init.git.git_status_lists import GitStatusLists


def get_status_lists(
    include_untracked: bool = False,
) -> GitStatusLists:
    """ Executes the Complete Git Status into File Change Operation.

**Parameters:**
 - include_untracked (bool): Whether to include untracked files in Git Status.

**Returns:**
 list[FileChange] - The List of FileChange information from Git Status.
    """
    return status_reader.read_git_status_output(
        status_runner.run_git_status() if not include_untracked else status_runner.run_untracked_status()
    )


def generate_file_changes(
    include_untracked: bool,
) -> Generator[FileChange, None, None]:
    """ Initialize FileChanges with a Generator.

**Parameters:**
 - include_untracked: Whether to include untracked files in the output.

**Returns:**
 Generator[FileChange] - Those precious File Changes.
    """
    status_lists = get_status_lists(include_untracked)
    yield from status_change_mapping.map_file_status_to_changes(
        status_lists.merge_all() if include_untracked else status_lists.merge_tracked()
    )


""" Setup Package Configuration
"""
from setuptools import setup, find_packages


setup(
    name="changelist-init",
    version="0.2",
	author='DK96-OS',
	description='Initialize and update Changelists information.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/DK96-OS/changelist-init/",
    project_urls={
        "Issues": "https://github.com/DK96-OS/changelist-init/issues",
        "Source Code": "https://github.com/DK96-OS/changelist-init/"
    },
	license='GPLv3',
    packages=find_packages(exclude=['test', 'test.*']),
    install_requires=[
        'changelist-data >= 0.2.3, < 0.3',
    ],
    entry_points={
        'console_scripts': [
            'changelist-init=changelist_init.__main__:main',
            'cl-init=changelist_init.__main__:main',
            'cl_init=changelist_init.__main__:main',
        ],
    },
    python_requires='>=3.10',
    keywords=['changelist', 'git', 'vcs'],
    classifiers=[
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ]
)


""" The Git Status of a file.
"""
from dataclasses import dataclass

from changelist_init.git.git_tracking_status import GitTrackingStatus


@dataclass(frozen=True)
class GitFileStatus:
    """ The Status information of a single File.

    Fields:
    - code (str): The two character code describing the file status.
    - file_path (str): The String path of the file.
    """
    code: str
    file_path: str

    def get_tracking_status(self) -> GitTrackingStatus:
        """ Obtain the Tracking Status of this File.

        Returns:
        GitTrackingStatus - The Tracking Status of this File.
        """
        if len(self.code) != 2:
            exit("Status Code should be two char length.")
        match self.code:
            case '??':
                return GitTrackingStatus.UNTRACKED
        if self.code.startswith(' '):
            return GitTrackingStatus.UNSTAGED
        if self.code.endswith(' '):
            return GitTrackingStatus.STAGED
        # If both code points are non-empty, is partially_staged
        return GitTrackingStatus.PARTIAL_STAGE



""" Reader for the Git Status Output String.
"""
from changelist_init.git.git_file_status import GitFileStatus
from changelist_init.git.git_status_lists import GitStatusLists


def read_git_status_output(
    status_string: str,
) -> GitStatusLists:
    """ Read the Git Status Output String.

    Fields:
    - status_string (str): The output of Git Status operation.

    Returns:
    GitStatusList - An object containing organized Git Status output.
    """
    if not isinstance(status_string, str):
        raise TypeError("Must be a String!")
    status_lists = GitStatusLists()
    for f in status_string.splitlines():
        if (file_status := read_git_status_line(f)) is not None:
            status_lists.add_file_status(file_status)
        else:
            print(f"Skipped: ${f}")
    return status_lists


def read_git_status_line(
    file_status_line: str
) -> GitFileStatus | None:
    """ Read a line of output from Git Status, into a GitFileStatus object.
    """
    if not isinstance(file_status_line, str):
        return None
    if len(file_status_line.strip()) < 3:
        return None
    if file_status_line.endswith('/'):
        return None
    return GitFileStatus(
        code=file_status_line[:2],
        file_path=file_status_line[3:],
    )



#!/usr/bin/python
from sys import argv, path


def main(): # Have to import after appending parent dir to path
    import changelist_init
    input_data = changelist_init.input.validate_input(argv[1:])
    # Generate FileChange info from git, update ChangelistDataStorage object.
    if not changelist_init.init_storage(
        storage=input_data.storage,
        include_untracked=input_data.include_untracked,
    ):
        exit("Failed to Merge new FileChanges into Changelists.")
    # Write Changelist Data file
    if not input_data.storage.write_to_storage():
        exit("Failed to Write Changelist Data File!")


if __name__ == "__main__":
    from pathlib import Path
    # Get the directory of the current file (__file__ is the path to the script being executed)
    current_directory = Path(__file__).resolve().parent.parent
    # Add the directory to sys.path
    path.append(str(current_directory))
    main()